#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>

#include "dta.h"
#include "dta_priv.h"

char	*me;
int	verbose = 0;
int	gen_data = 0;

void	usage(void)
{
	fprintf(stderr,"Usage: %s [-drv] [-e expfile] [-n raw-data-count] [-p preload-exp] [-R num-bits-in-exp]\n",me);
}

big_int_t		modulus_data;
big_int_t const		*modulus = &modulus_data;
static int		secret_exp_set = 0;
static exp_t		secret_exp;
static exp_t		preload;

void	load_exp_from_file(char		*fn,
			   exp_t	*ep)
{
	FILE	*iob;

	if (NULL == (iob = fopen(fn,"r"))) {
		fprintf(stderr,"%s: cannot open key file %s\n",me,fn);
		usage();
		exit(1);
	}
	if (!load_exp(ep,iob)) {
		fprintf(stderr,"%s: illegal input in exponent file %s\n",me,fn);
		exit(1);
	}
	if (1 != ep->bits[0]) {
		fprintf(stderr,"%s: exponent's low order bit must be 1!\n",me);
		exit(1);
	}
	(void) fclose(iob);
}

/*
 * Uniformly random exponent bits.
 */
void	generate_random_exp(int	nbits)
{
	int	i;

	secret_exp.bits[0] = 1;
	for (i = 1; i < nbits; i++) {
		secret_exp.bits[i] = (random() & 1);
	}
	for (i = nbits; --i >= 0 && secret_exp.bits[i] == 0;)
		;
	secret_exp.length = i+1;
}

void	query_card(big_int_t		*w,
		   sim_time_t		*t,
		   big_int_t const	*x)
{
	modexp(w,t,x,&secret_exp,modulus);
}

int	main(int	ac,
	     char	**av)
{
	int	opt;
	int	count = 1000000;
	/* generating data for external simulation/analysis */
	exp_t	guess;
	int	nbits = MAXBITS;

	me = strrchr(av[0],'/');
	if (!me) me = av[0]; else ++me;

	srandom(time((time_t *) 0));
	while (EOF != (opt = getopt(ac,av,"de:n:p:rR:s:v"))) switch (opt) {
	case 'd':
		++gen_data;
		break;
	case 'e':
		load_exp_from_file(optarg,&secret_exp);
		secret_exp_set = 1;
		break;
	case 'n':
		count = atoi(optarg);
		break;
	case 'p':
		/*
		 * allows preloading of low order bits of exponent
		 * to permit testing of later stages of detection algorithm
		 */
		load_exp_from_file(optarg,&preload);
		dta_preload(&preload);
		break;
	case 'r':
		generate_random_exp(nbits);
		secret_exp_set = 1;
		break;
	case 'R':
		nbits = atoi(optarg);
		if (nbits > MAXBITS) {
			fprintf(stderr,"%s: -R specifies too many bits!\n",me);
			usage();
			exit(1);
		}
		generate_random_exp(nbits);
		secret_exp_set = 1;
		break;
	case 's':
		srandom(atoi(optarg));
		/* allow user to set seed -- if -r/R used, should follow -s */
		break;
	case 'v':
		++verbose;
		break;
	default:
		usage();
		exit(1);
	}

	if (0) {
		big_int_t	x,y,z1,z2;
		sim_time_t	t1,t2;

		bigrand(&x); bigrand(&y);
		modmult(&z1,&t1, &x,&y,modulus);
		modmult(&z2,&t2, &x,&y,modulus);
		if (big_int_equal(&z1,&z2)) {
			fprintf(stderr,"mod mult values okay.\n");
		} else {
			fprintf(stderr,"mod mult values BAD.\n");
			abort();
		}
		if (t1 == t2) {
			fprintf(stderr,"mod mult times okay.\n");
		} else {
			fprintf(stderr,"mod mult times BAD.\n");
			abort();
		}
		modexp(&z1,&t1, &x,&secret_exp,modulus);
		modexp(&z2,&t2, &x,&secret_exp,modulus);
		if (big_int_equal(&z1,&z2)) {
			fprintf(stderr,"mod exp values okay.\n");
		} else {
			fprintf(stderr,"mod exp values BAD.\n");
			abort();
		}
		if (t1 == t2) {
			fprintf(stderr,"mod exp time okay.\n");
		} else {
			fprintf(stderr,"mod exp time BAD.\n");
			abort();
		}
	}

	if (!secret_exp_set) {
		fprintf(stderr,"%s: no exponent chosen\n",me);
		usage();
		exit(1);
	}

	if (gen_data) {
		int	 	i;
		big_int_t	x;

		for (i = 0; i < count; i++) {
			bigrand(&x);
			modexp_data(&x,&secret_exp,modulus);
			putchar('\n');
		}
		return 0;
	}

	if (verbose) {
		printf("Mounting DTA attack on secret exponent\n");
		print_exp(&secret_exp,stdout);
		putchar('\n');
	}

	dta(&guess);

	printf("DTA output:\n");
	print_exp(&guess,stdout);
	printf("\nCorrect secret exponent:\n");
	print_exp(&secret_exp,stdout);
	putchar('\n');

	return 0;
}
